import 'package:meta/meta.dart';
import 'dfbaserepository.dart';
import '../model/dfwebmodel.dart';
import '../control/dfresponse.dart';


import '../control/dfquery.dart';
import '../entity/dfwebentity.dart';

import 'package:nui_web/nui_web.dart';
import 'package:http/http.dart' as http;
import 'package:ssl_pinning_plugin/ssl_pinning_plugin.dart';

class InvalidWebReqException extends DfError {
  InvalidWebReqException() : super(code: 8100, message: "Invalid web request. Please specify web request in EntityJoin");
}

class SslPinningFailException extends DfError {
  SslPinningFailException() : super(code: 8101, message: "Fail to identify the backend server. SSL Pinning failure.");
}

class DfWebRepository extends DfBaseRepository<DfWebModel> {

  DfWebRepository({@required DfWebModel model}) : super(model: model);

  DfWebEntity findWebEntity(DfQuery query) {
    var baseEntity = query.entityJoin.entities[0];
    return model.entities.where((entity) => entity.name == baseEntity.name).toList()[0];
  }

  @override
  Future<List<DfResponse>> delete(List<DfQuery> queries) async {
    List<DfResponse> responses = List<DfResponse>();
    for (DfQuery query in queries) {
      if (query.entityJoin != null && query.entityJoin.entities.length > 0) {
        DfWebEntity daoWebSubContainer = findWebEntity(query);
        Map<String,String> headers = formFinalHeaders(daoWebSubContainer.headers);
        try {
          await _checkSslPinning(headers);
          NUIWeb web = await buildWebRef(headers);
          DfWebReqParam reqParam = daoWebSubContainer.convertToWebReqParam(query);
          if (reqParam.pathVariables == null || reqParam.pathVariables.length == 0) {
            responses.add(DfResponse(error: InvalidWebReqException()));
          }
          else {
            DfResponse response = DfResponse();
            String url = daoWebSubContainer.formUrl(query);
            response.dataFromHttpResponse( await web.doDELETE(url: url));
            responses.add(response);
          }        
        }
        catch (e,stacktrace) {
          print("Delete got error: $e, $stacktrace so set the response.error");
          responses.add(DfResponse(error: e));
        }
      }
      else {
        responses.add(DfResponse(error: InvalidWebReqException()));
      }
    }  
    return responses;
  }

  _checkSslPinning(Map<String,dynamic> headers) async {
    var sslPinningCri = model.sslPinningCri;
    if (sslPinningCri != null) {
        try {
          String checkMsg = await SslPinningPlugin.check(
            serverURL: model.serverUrl,
            headerHttp: headers == null ? Map() : headers,
            httpMethod: HttpMethod.Head,
            sha: sslPinningCri.shatype == SslPinningCriShaType.SHA1 ? SHA.SHA1 : SHA.SHA256,
            allowedSHAFingerprints: sslPinningCri.shakeys,
            timeout: 60
          );
        }
        catch (e) {
          print("checkSslPinning() error: ${e.toString()} occured.");
          throw SslPinningFailException();
        }
    }   
  } 
  
    @override
    Future<DfResponse> fetch(DfQuery query) async {
      DfResponse response = DfResponse();

      if (query.entityJoin != null && query.entityJoin.entities.length > 0) {
        DfWebEntity daoWebSubContainer = findWebEntity(query);
        Map<String,String> headers = formFinalHeaders(daoWebSubContainer.headers);
        try {
          await _checkSslPinning(headers);
          NUIWeb web = await buildWebRef(headers);
          DfWebReqParam reqParam = daoWebSubContainer.convertToWebReqParam(query);

          String url = daoWebSubContainer.formUrl(query);
          
          print("nui doGET WebAdapter Url: ${model.serverUrl}$url");
          
          //Plain connection
          // url = "$url?";
          // reqParam.reqParams.forEach((k,v) => url = "$url$k=$v&");
          // http.Response plainHttpResp = await http.get("${model.serverUrl}$url",headers: headers);
          // response.dataFromPlainHttpResponse(plainHttpResp);
          
          //Nui
          NUIHttpResponse nuiResp = await NUIWeb.get(module: model.name).doGET(url: url, params: reqParam.reqParams);
          response.dataFromHttpResponse(nuiResp);
          
          print("finish doGET WebAdapter Url: ${model.serverUrl}$url");
        }
        catch (e,stacktrace) {
          print("Fetch got error: $e, $stacktrace so set the response.error");
          response.error = e;
        }
      }
      else {
        response.error = InvalidWebReqException();
      }
      return response;
    }
  
    @override
    Future<List<DfResponse>> save(List<DfQuery> queries) async {
      List<DfResponse> responses = List<DfResponse>();

      for (DfQuery query in queries) {
        if (query.entityJoin != null && query.entityJoin.entities.length > 0) {
          DfWebEntity daoWebSubContainer = findWebEntity(query);
          try {
            await _checkSslPinning(daoWebSubContainer.headers);
            Map<String,String> headers = formFinalHeaders(daoWebSubContainer.headers);
            NUIWeb web = await buildWebRef(headers);
            DfWebReqParam reqParam = daoWebSubContainer.convertToWebReqParam(query);

            String url = daoWebSubContainer.formUrl(query);
            print("WebAdapter Url: $url");
            DfResponse response = DfResponse();
            if (query.purpose == DfQueryPurpose.ADD ) {
              response.dataFromHttpResponse( await web.doPOST(url: url, body: reqParam.reqBody) );          
            }
            else if (query.purpose == DfQueryPurpose.UPDATE_WHOLE_ENTITY_ATTRS) {
              response.dataFromHttpResponse( await web.doPUT(url: url, body: reqParam.reqBody) );            
            }
            else if (query.purpose == DfQueryPurpose.UPDATE_SOME_RNTITY_ATTRS) {
              response.dataFromHttpResponse( await web.doPATCH(url: url, body: reqParam.reqBody) );            
            }
            responses.add(response);
          }
        catch (e,stacktrace) {
            print("Save got error: $e, $stacktrace so set the response.error");
            responses.add(DfResponse(error: e));
          }
        }
        else {
          responses.add(DfResponse(error: InvalidWebReqException()));
        }
      }
      return responses;
    }

  @override
  Future<void> startInit() async {
  }

  Map<String,String> formFinalHeaders(Map<String,String> headers) {
    Map<String,String> allHeaders = new Map<String,String>();
    if (headers != null)
      allHeaders.addAll(headers);
    if (model.commonHeaders != null)
      allHeaders.addAll(model.commonHeaders);
    return allHeaders;
  }

  Future<NUIWeb> buildWebRef(Map<String,String> headers) async {
    var builder = NUIWebBuilder();
    builder.module(model.name);
    builder.baseURL(model.serverUrl);
    builder.headers(() => headers);
    return await builder.build();    
  }  

  @override
  Future<void> stop() async {
  }
}